# -*- coding: utf-8 -*-
##############################################################################
#
#    wms module for OpenERP, WMS Extended for Open ERP
#    Copyright (C) 2011 SYLEAM Info Services (<http://www.Syleam.fr/>)
#              Sylvain Garancher <sylvain.garancher@syleam.fr>
#
#    This file is a part of wms
#
#    wms is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    wms is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

import netsvc
from osv import osv
from osv import fields
from tools.translate import _


class stock_burst(osv.osv):
    _name = 'stock.burst'
    _description = 'Burst picking'
    _rec_name = 'stock_picking_id'

    _columns = {
        'stock_picking_id': fields.many2one('stock.picking', 'Picking'),
        'picking_ids': fields.many2many('stock.picking', 'rel_burst_picking', 'burst_id', 'picking_id', 'Backorders', ),
        'line_ids': fields.one2many('stock.burst.line', 'burst_id', 'Lines', ),
        'state': fields.selection([('draft', 'Draft'), ('done', 'Done')], 'State', ),
        'type': fields.selection([('in', 'Receipt'), ('out', 'Picking')], 'Type', ),
    }

    _defaults = {
        'state': lambda *a: 'draft',
    }

    def copy(self, cr, uid, id, default=None, context=None):
        """
        When duplicate data, we need to erase line to original picking and picking relation
        """
        if context is None:
            context = {}

        default['stock_picking_id'] = False
        default['picking_ids'] = []

        return super(stock_burst, self).copy(cr, uid, id, default, context=context)

    def _merge(self, cr, uid, id, context=None):
        """
        Merge burst lines
        Returns burst lines in dico : {'product_key': (total_qty, [(qty, line_id, remark_id), (qty, line_id, remark_id), ])}
        """
        if context is None:
            context = {}

        # Retrieve burst lines
        burst_line_obj = self.pool.get('stock.burst.line')
        burst_line_ids = burst_line_obj.search(cr, uid, [('burst_id', '=', id), ('is_done', '=', False)], context=context)
        burst_lines = burst_line_obj.browse(cr, uid, burst_line_ids, context=context)

        known_lines = {}
        lines_to_delete = []
        burst_type = self.read(cr, uid, id, ['type'], context=context)['type']
        for line in burst_lines:
            product_key = "%d_%d_%d" % (line.product_id.id, line.prodlot_id and line.prodlot_id.id or 0,
                                        line.tracking_id and line.tracking_id.id or 0)
            # Remarks are not significant for type 'out'
            rem_id = (burst_type != 'out') and line.remark_id.id or 0
            # Check if product_id is known
            if product_key in known_lines.keys():
                # Check if product is known with same remark_id (or 0 for no remark)
                if rem_id in known_lines[product_key].keys():
                    # Add quantities
                    known_lines[product_key][rem_id].quantity += line.quantity
                    lines_to_delete.append(line.id)
                else:
                    # Add remark_id in dico
                    known_lines[product_key][rem_id] = line
            else:
                # Creates dico for this product_key
                known_lines[product_key] = {rem_id: line}

        # Saves modified lines
        for p, lines in known_lines.items():
            # Total quantity for this product
            total = 0
            # List lines for this product (one line per remark_id)
            lines_list = []
            for rem_id, line in lines.items():
                burst_line_obj.write(cr, uid, [line.id], {'quantity': line.quantity}, context=context)
                total += line.quantity
                lines_list.append((line.quantity, line.id, rem_id))
            # final format :
            known_lines[p] = (total, lines_list)

        # Delete lines to delete
        if lines_to_delete:
            burst_line_obj.unlink(cr, uid, lines_to_delete, context=context)

        return known_lines

    def _merge_and_compare(self, cr, uid, ids, context=None):
        """
        Merge burst lines and compare them with move lines of picking
        """
        if context is None:
            context = {}

        # Retrieve merged burst lines
        burst_line_obj = self.pool.get('stock.burst.line')
        burst_lines = self._merge(cr, uid, ids[0], context=context)
        # No new line : quit
        if not burst_lines:
            return

        # Retrieve merged move lines
        picking_obj = self.pool.get('stock.picking')
        move_obj = self.pool.get('stock.move')
        picking_id = self.read(cr, uid, ids, ['stock_picking_id'], context=context)[0]['stock_picking_id'][0]
        backorder_ids = self.pool.get('stock.picking').search(cr, uid, [('backorder_id', '=', picking_id), ('state', 'not in', ('done', 'cancel', 'draft'))], context=context)
        if backorder_ids:
            picking = self.pool.get('stock.picking').browse(cr, uid, backorder_ids[0], context=context)
        else:
            picking = self.pool.get('stock.picking').browse(cr, uid, picking_id, context=context)
        picking_moves = picking_obj.move_merge(cr, uid, picking.id, context=context)

        # Compare burst lines and move lines
        done_burst_lines = []
        done_move_lines = []
        missing = []
        for p, lines in burst_lines.items():
            # Unknown product, create moves
            if not p in picking_moves.keys():
                for line in lines[1]:
                    (product_id, prodlot_id, tracking_id) = p.split('_')
                    product_id = int(product_id)
                    product = self.pool.get('product.product').browse(cr, uid, product_id, context=context)
                    if picking.type == 'in' and not (picking.address_id.partner_id and picking.address_id.partner_id.property_stock_supplier):
                        raise osv.except_osv(_('Error'), _('The stock location on the partner is not well configure!'))

                    if picking.type == 'out':
                        location_id = picking.warehouse_id.lot_input_id.id
                        location_dest_id = picking.address_id.partner_id.property_stock_customer.id
                        location_dest = picking.address_id.partner_id.property_stock_customer
                    else:
                        location_id = picking.address_id.partner_id.property_stock_supplier.id
                        location_dest_id = picking.warehouse_id.lot_input_id.id
                        location_dest = picking.warehouse_id.lot_input_id

                    # check if the location is chained
                    if product.location_type in ('fixed', 'crossdock'):
                        p_dest_id, p_auto, p_delay = self.pool.get('stock.location').chained_location_get(cr, uid, location_dest, product=product)
                        if p_dest_id and p_auto == 'transparent':
                            location_dest_id = p_dest_id.id

                    move_id = move_obj.create(cr, uid, {'name': product.name,
                                              'location_id': location_id,
                                              'location_dest_id': location_dest_id,
                                              'product_id': int(product_id),
                                              'product_uom': product.uom_id.id,
                                              'prodlot_id': int(prodlot_id),
                                              'tracking_id': int(tracking_id),
                                              'product_qty': line[0],
                                              'picking_id': picking.id,
                                              'remark_id': (line[2] or False),
                                              'comment': _("Not in initial receipt"),
                                             }, context=context)
                    done_burst_lines.append(line[1])
                    done_move_lines.append(move_id)
            else:
                move = picking_moves[p]
                no_remark = [t for t in lines[1] if t[2] == 0]
                # Quantity is superior : split picking line
                if len(no_remark) and move[0] < no_remark[0][0]:
                    lines = (lines[0], [((t[2] == 0) and (move[0], t[1], t[2]) or t) for t in lines[1]])
                    lines[1].append((no_remark[0][0] - move[0], no_remark[0][1], no_remark[0][2], _("Not in initial receipt")))

                # Save first line
                move_obj.write(cr, uid, [move[1]], {'product_qty': lines[1][0][0], 'remark_id': (lines[1][0][2] or False)}, context=context)
                done_burst_lines.append(lines[1][0][1])
                done_move_lines.append(move[1])
                # Save all other lines with copies of move
                for line in lines[1][1:]:
                    if line[0] > 0:
                        move_id = move_obj.copy(cr, uid, move[1], {'product_qty': line[0],
                                                         'remark_id': (line[2] or False),
                                                         'comment': ((len(line) > 3) and line[3] or False),
                                                        }, context=context)
                        done_burst_lines.append(line[1])
                        done_move_lines.append(move_id)

                # Quantity is inferior : missing products
                if picking_moves[p][0] > lines[0]:
                    missing.append((move[1], move[0] - lines[0]))

        # Check for products not found in burst
        for p, (qty, move_id) in picking_moves.items():
            if p not in burst_lines.keys():
                # Set quantity to 0
                move_obj.write(cr, uid, [move_id], {'state': 'done', 'product_qty': 0}, context=context)
                # Add to missing list
                missing.append((move_id, qty))

        burst_type = self.read(cr, uid, ids, ['type'], context=context)[0]['type']
        wf_service = netsvc.LocalService("workflow")
        del_empty_move = []
        if len(missing) > 0:
            no_keep = (picking.type == 'in' and not picking.warehouse_id.keep_missing_product_in) or False
            # Generate new picking for missing products
            new_picking = picking_obj.copy(cr, uid, picking_id,
                    {
                        'name': self.pool.get('ir.sequence').get(cr, uid, 'stock.picking'),
                        'move_lines': [],
                        'state': 'draft',
                        'backorder_id': picking_id,
                    })

            for (move_id, qty) in missing:
                # Generate move lines in new_picking for missing products
                move_obj.copy(cr, uid, move_id, {'product_qty': qty, 'picking_id': new_picking, 'remark_id': False})
                if no_keep:
                    del_empty_move.append(move_id)

            if del_empty_move:
                #move_obj.write(cr, uid, del_empty_move, {'state': 'draft'})
                move_obj.unlink(cr, uid, del_empty_move)

            # Sets state to assigned for new picking
            wf_service.trg_validate(uid, 'stock.picking', new_picking, 'button_confirm', cr)
            if burst_type != 'out':
                picking_obj.force_assign(cr, uid, [new_picking])

        for val in done_move_lines:
            if val in del_empty_move:
                done_move_lines.remove(val)

        if burst_type == 'in':
            # Set state to Done for picking
            if done_move_lines:
                move_obj.write(cr, uid, done_move_lines, {'state': 'done'}, context=context)

            picking_obj.force_assign(cr, uid, [picking.id])

        elif burst_type == 'out':
            picking_obj.action_move(cr, uid, [picking.id])

        wf_service.trg_validate(uid, 'stock.picking', picking.id, 'button_done', cr)

        # Set is_done=True for finished burst lines
        burst_line_obj.write(cr, uid, done_burst_lines, {'is_done': True}, context=context)

        # Add picking in M2M
        self.write(cr, uid, ids, {'picking_ids': [(4, picking.id)]}, context=context)

    def valid_partial(self, cr, uid, ids, context=None):
        """ Validate partial picking """
        if context is None:
            context = {}

        self._merge_and_compare(cr, uid, ids, context=context)

    def valid_done(self, cr, uid, ids, context=None):
        """ Validate full picking """
        if context is None:
            context = {}

        self._merge_and_compare(cr, uid, ids, context=context)
        self.write(cr, uid, ids, {'state': 'done'}, context=context)

stock_burst()


class stock_burst_line(osv.osv):
    _name = 'stock.burst.line'
    _description = 'Burst picking line'
    _rec_name = 'product_id'

    _columns = {
        'product_id': fields.many2one('product.product', 'Product', required=True, ),
        'quantity': fields.float('Quantity', required=True, ),
        'remark_id': fields.many2one('stock.move.remark', 'Observation', ),
        'is_done': fields.boolean('Done', ),
        'burst_id': fields.many2one('stock.burst', 'Burst', ),
        'prodlot_id': fields.many2one('stock.production.lot', 'Production lot', ),
        'tracking_id': fields.many2one('stock.tracking', 'Tracking', ),
    }

    _defaults = {
        'is_done': lambda *a: False,
        'quantity': lambda *a: 1,
    }

stock_burst_line()


# vim:expandtab:smartindent:tabstop=4:softtabstop=4:shiftwidth=4:
